# Analysis

##################################################
# Assembly quality

rule analysis_quast:
    input:
        os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta")
    output:
        os.path.join(RESULTS_DIR, "analysis/quast/{rtype}/{tool}/report.tsv")
    log:
        out="logs/quast.{rtype}.{tool}.out.log",
        err="logs/quast.{rtype}.{tool}.err.log"
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads:
        config["quast"]["threads"]
    conda:
        os.path.join(ENV_DIR, "quast.yaml")
    message:
        "Assess assembly quality w/ QUAST: {input}"
    shell:
        "(date && metaquast.py --max-ref-number 0 --threads {threads} {input} -o $(dirname {output}) && date) 2> {log.err} > {log.out}"

##################################################
# Proteins

# for CD-HIT
rule analysis_bbmap_rename:
    input:
        os.path.join(RESULTS_DIR, "annotation/prodigal/{rtype}/{tool}/proteins.faa")
    output:
        temp(os.path.join(RESULTS_DIR, "analysis/cdhit/{rtype}_{tool}.faa"))
    log:
        out="logs/analysis_bbmap_rename_{rtype}.{tool}.out.log",
        err="logs/analysis_bbmap_rename_{rtype}.{tool}.err.log"
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads:
        config["bbmap"]["threads"]
    params:
        # TODO: is this prefix okay?
        prefix=lambda wildcards: "%s_%s" % (wildcards.rtype, wildcards.tool)
    conda:
        os.path.join(ENV_DIR, "bbmap.yaml")
    message:
        "BBMAP: rename FASTA entries in {input}"
    shell:
        "(date && rename.sh in={input} out={output} prefix={params.prefix} ignorejunk=t && date) 2> {log.err} > {log.out}"

rule analysis_cdhit:
    input:
        faa1=os.path.join(RESULTS_DIR, "analysis/cdhit/{rtype1}_{tool1}.faa"),
        faa2=os.path.join(RESULTS_DIR, "analysis/cdhit/{rtype2}_{tool2}.faa")
    output:
        faa12=os.path.join(RESULTS_DIR, "analysis/cdhit/{rtype1}_{tool1}__{rtype2}_{tool2}.faa"),
        faa21=os.path.join(RESULTS_DIR, "analysis/cdhit/{rtype2}_{tool2}__{rtype1}_{tool1}.faa")
    log:
        out="logs/analysis_cdhit.{rtype1}.{tool1}.{rtype2}.{tool2}.out.log",
        err="logs/analysis_cdhit.{rtype1}.{tool1}.{rtype2}.{tool2}.err.log"
    wildcard_constraints:
        rtype1="|".join(READ_TYPES),
        rtype2="|".join(READ_TYPES),
        tool1="|".join(ASSEMBLERS),
        tool2="|".join(ASSEMBLERS)
    threads: 1 # TODO
    conda:
        os.path.join(ENV_DIR, "cdhit.yaml")
    message:
        "CD-HIT: {input}"
    shell:
        "(date && "
        "cd-hit-2d -i {input.faa1} -i2 {input.faa2} -o {output.faa12} -c 0.9 -n 5 -d 0 -M 16000 -T 8 && "
        "cd-hit-2d -i {input.faa2} -i2 {input.faa1} -o {output.faa21} -c 0.9 -n 5 -d 0 -M 16000 -T 8 && "
        "date) 2> {log.err} > {log.out}"

##################################################
# Circular contigs

rule circ_contigs_fasta:
    input:  
        os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta")
    output: 
        split1=temp(os.path.join(RESULTS_DIR, "analysis/circ_blast/{rtype}/{tool}/ASSEMBLY.split.1.fasta")),
        split2=temp(os.path.join(RESULTS_DIR, "analysis/circ_blast/{rtype}/{tool}/ASSEMBLY.split.2.fasta"))
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    conda:
        os.path.join(ENV_DIR, "rgi.yaml") # NOTE: need an env. containing Biopython
    message:
        "Split sequences for BLAST-based analysis of circular contigs: {input}"
    script:
        os.path.join(SRC_DIR, "circ_contigs_fasta.py")

rule circ_contigs_makeblastdb:
    input:
        os.path.join(RESULTS_DIR, "analysis/circ_blast/{rtype}/{tool}/ASSEMBLY.split.2.fasta")
    output:
        expand(
            os.path.join(RESULTS_DIR, "analysis/circ_blast/{{rtype}}/{{tool}}/ASSEMBLY.split.2.fasta.{ext}"),
            ext=["nhr", "nin", "nsq"]
        )
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    params:
        db=lambda wildcards, output: os.path.splitext(output[0])[0]
    conda:
        os.path.join(ENV_DIR, "rgi.yaml") # NOTE: need an env. containing BLAST
    shell: 
        "makeblastdb -dbtype nucl -input_type fasta -in {input} -out {params.db}"

rule circ_contigs_blastn:
    input:
        db=expand(
            os.path.join(RESULTS_DIR, "analysis/circ_blast/{{rtype}}/{{tool}}/ASSEMBLY.split.2.fasta.{ext}"),
            ext=["nhr", "nin", "nsq"]
        ),
        fasta=os.path.join(RESULTS_DIR, "analysis/circ_blast/{rtype}/{tool}/ASSEMBLY.split.1.fasta")
    output:
        os.path.join(RESULTS_DIR, "analysis/circ_blast/{rtype}/{tool}/ASSEMBLY.tsv")
    log:
        out="logs/circ_blast.{rtype}.{tool}.out.log",
        err="logs/circ_blast.{rtype}.{tool}.err.log"
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    params:
        db=lambda wildcards, input: os.path.splitext(input["db"][0])[0],
        outfmt="-outfmt '6 qseqid sseqid qlen slen qstart qend sstart send evalue bitscore score length pident nident mismatch gapopen gaps qcovhsp'"
    conda:
        os.path.join(ENV_DIR, "rgi.yaml") # NOTE: need an env. containing BLAST
    shell: 
        "(date && "
        "blastn -task megablast -num_alignments 50 -perc_identity 95 {params.outfmt} -query {input.fasta} -db {params.db} -out {output} && "
        "date) 2> {log.err} > {log.out}"

rule circ_contigs_filter:
    input:
        os.path.join(RESULTS_DIR, "analysis/circ_blast/{rtype}/{tool}/ASSEMBLY.tsv")
    output:
        circ=os.path.join(RESULTS_DIR, "analysis/circ_blast/{rtype}/{tool}/ASSEMBLY.circ.tsv"),
        compl=os.path.join(RESULTS_DIR, "analysis/circ_blast/{rtype}/{tool}/ASSEMBLY.compl.tsv")
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    run:
        import pandas
        cols = "qseqid sseqid qlen slen qstart qend sstart send evalue bitscore score length pident nident mismatch gapopen gaps qcovhsp".split(" ")
        df = pandas.read_csv(input[0], sep="\t", header=None, names=cols)
        # contig ID and split ID
        df = df.assign(qcontigid=df.qseqid.apply(lambda x: re.sub("_(1|2)$", "", x)))
        df = df.assign(qsplitid=df.qseqid.apply(lambda x: x.split("_")[-1]))
        df = df.assign(scontigid=df.sseqid.apply(lambda x: re.sub("_(1|2)$", "", x)))
        df = df.assign(ssplitid=df.sseqid.apply(lambda x: x.split("_")[-1]))
        # filter by IDs
        df = df[ (df.qcontigid == df.scontigid) & (df.qsplitid != df.ssplitid) ]
        # filter by mismatches/gaps
        df = df[ (df.mismatch <= 5) & (df.gaps <= 5) ]
        # putative circ. contigs: filter by alignment start/end
        df_circ = df[ (df.qstart == 1) & (df.send == df.slen) ]
        # weird hits
        df_compl = df[ (df.qstart == 1) & (df.qend == df.qlen) & (df.sstart == df.slen) & (df.send == 1) ]
        # save
        df_circ.to_csv(output.circ, sep="\t", header=True, index=False, index_label=False)
        df_compl.to_csv(output.compl, sep="\t", header=True, index=False, index_label=False)

##################################################
# Mash

# Assemblies
rule mash_sketch:
    input:
        os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta")
    output:
        os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.msh")
    log:
        out="logs/mash.sketch.{rtype}.{tool}.out.log",
        err="logs/mash.sketch.{rtype}.{tool}.err.log"
    threads:
        1
    conda:
        os.path.join(ENV_DIR, "mash.yaml")
    shell:
        "(date && ofile={output} && mash sketch -k 31 -s 10000 -S 42 -o ${{ofile%.*}} {input} && date) 2> {log.err} > {log.out}"

rule mash_sketch_paste:
    input:
        lr=expand(os.path.join(RESULTS_DIR, "assembly/lr/{tool}/ASSEMBLY.msh"), tool=config["assemblers"]["lr"]),
        sr=expand(os.path.join(RESULTS_DIR, "assembly/sr/{tool}/ASSEMBLY.msh"), tool=config["assemblers"]["sr"]),
        hy=expand(os.path.join(RESULTS_DIR, "assembly/hy/{tool}/ASSEMBLY.msh"), tool=config["assemblers"]["hy"])
    output:
        os.path.join(RESULTS_DIR, "analysis/mash/contigs.msh")
    threads:
        1
    conda:
        os.path.join(ENV_DIR, "mash.yaml")
    shell:
        "ofile={output} && mash paste ${{ofile%.*}} {input}"
    
rule mash_dist:
    input:
        os.path.join(RESULTS_DIR, "analysis/mash/contigs.msh")
    output:
        os.path.join(RESULTS_DIR, "analysis/mash/contigs.dist")
    log:
        out="logs/mash.dist.contigs.out.log",
        err="logs/mash.dist.contigs.err.log"
    threads:
        config["mash"]["threads"]
    conda:
        os.path.join(ENV_DIR, "mash.yaml")
    shell:
        "(date && "
        "mash dist -t -p {threads} {input} {input} > {output} && "
        "date) 2> {log.err} > {log.out}"

# Contigs
rule mash_sketch_contigs:
    input:
        os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta")
    output:
        os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.contigs.msh")
    log:
        out="logs/mash.sketch.{rtype}.{tool}.contigs.out.log",
        err="logs/mash.sketch.{rtype}.{tool}.contigs.err.log"
    threads:
        1
    conda:
        os.path.join(ENV_DIR, "mash.yaml")
    shell:
        "(date && ofile={output} && mash sketch -i -k 31 -s 1000 -S 42 -o ${{ofile%.*}} {input} && date) 2> {log.err} > {log.out}"

rule mash_screen_contigs:
    input:
        fasta=expand(os.path.join(RESULTS_DIR, "assembly/{rtype_tool}/ASSEMBLY.fasta"), rtype_tool=["%s/%s" % (rtype, tool) for rtype, tool in READ_ASSEMBLERS]),
        msh=expand(os.path.join(RESULTS_DIR, "assembly/{rtype_tool}/ASSEMBLY.contigs.msh"), rtype_tool=["%s/%s" % (rtype, tool) for rtype, tool in READ_ASSEMBLERS])
    output:
        expand(os.path.join(RESULTS_DIR, "analysis/mash/screen.{rtype_tool}.tsv"), rtype_tool=["%s.%s" % (rtype, tool) for rtype, tool in READ_ASSEMBLERS])
    threads:
        config["mash"]["threads"]
    conda:
        os.path.join(ENV_DIR, "mash.yaml")
    script:
        os.path.join(SRC_DIR, "mash_screen_contigs.py")

##################################################
# MMseqs2

# rule analysis_mmseqs2_db:
#     input:
#         os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta")
#     output:
#         os.path.join(RESULTS_DIR, "analysis/mmseqs2/{rtype}/{tool}/ASSEMBLY_db")
#     log:
#         out="logs/mmseqs2_db.{rtype}.{tool}.out.log",
#         err="logs/mmseqs2_db.{rtype}.{tool}.err.log"
#     wildcard_constraints:
#         rtype="|".join(READ_TYPES),
#         tool="|".join(ASSEMBLERS)
#     conda:
#         os.path.join(ENV_DIR, "mmseqs2.yaml")
#     message:
#         "Create MMseqs2 DB from {input}"
#     shell:
#         "(date && mmseqs createdb {input} {output} && date) 2> {log.err} > {log.out}"

# rule analysis_mmseqs2_compare:
#     input:
#         db1=os.path.join(RESULTS_DIR, "analysis/mmseqs2/{rtype1}/{tool1}/ASSEMBLY_db"),
#         db2=os.path.join(RESULTS_DIR, "analysis/mmseqs2/{rtype2}/{tool2}/ASSEMBLY_db")
#     output:
#         os.path.join(RESULTS_DIR, "analysis/mmseqs2/comparison/{rtype1}_{tool1}__{rtype2}_{tool2}")
#     log:
#         out="logs/mmseqs2.{rtype1}.{tool1}.{rtype2}.{tool2}.out.log",
#         err="logs/mmseqs2.{rtype1}.{tool1}.{rtype2}.{tool2}.err.log"
#     wildcard_constraints:
#         rtype1="|".join(READ_TYPES),
#         rtype2="|".join(READ_TYPES),
#         tool1="|".join(ASSEMBLERS),
#         tool2="|".join(ASSEMBLERS)
#     threads:
#         config["mmseqs2"]["threads"]
#     conda:
#         os.path.join(ENV_DIR, "cd-hit.yaml")
#     message:
#         "Create MMseqs2 compare: {input}"
#     shell:
#         # TODO: "mmseqs2_tmp" ??? (see old files)
#         "(date && mmseqs rbh {input.db1} {input.db2} {output} --min-seq-id 0.9 --threads {threads} && date) 2> {log.err} > {log.out}"

# rule analysis_mmseqs2_m8_convert:
#     input:
#         db1=os.path.join(RESULTS_DIR, "analysis/mmseqs2/{rtype1}/{tool1}/ASSEMBLY_db"),
#         db2=os.path.join(RESULTS_DIR, "analysis/mmseqs2/{rtype2}/{tool2}/ASSEMBLY_db"),
#         rbh=os.path.join(RESULTS_DIR, "analysis/mmseqs2/comparison/{rtype1}_{tool1}__{rtype2}_{tool2}")
#     output:
#         os.path.join(RESULTS_DIR, "analysis/mmseqs2/comparison/{rtype1}_{tool1}__{rtype2}_{tool2}.m8")
#     log:
#         out="logs/mmseqs2_convert.{rtype1}.{tool1}.{rtype2}.{tool2}.out.log",
#         err="logs/mmseqs2_convert.{rtype1}.{tool1}.{rtype2}.{tool2}.err.log"
#     conda:
#         os.path.join(ENV_DIR, "cd-hit.yaml")
#     message:
#         "Create MMseqs2 compare: {input}"
#     shell:
#         "(date && mmseqs convertalis {input.db1} {input.db2} {input.rbh} {output} && date) 2> {log.err} > {log.out}"
