# Assembly

# Long reads
rule assembly_lr_flye:
    input:
        os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.fastq.gz")
    output:
        # NOTE: NOT the name for "final" FASTA file which will be created AFTER polishing
        os.path.join(RESULTS_DIR, "assembly/lr/flye/assembly.fasta")
    threads:
        config["flye"]["threads"]
    log:
        out="logs/flye.out.log",
        err="logs/flye.err.log"
    conda:
        os.path.join(ENV_DIR, "flye_v2_7.yaml")
    message:
        "Assembly: long reads: Flye"
    shell:
        "(date && flye --nano-raw {input} --meta --out-dir $(dirname {output}) --genome-size {config[flye][genome_size]} --threads {threads} && "
        "date) 2> {log.err} > {log.out}"

rule polishing_lr_racon:
    input:
        lr=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.fastq.gz"),
        asm=os.path.join(RESULTS_DIR, "assembly/lr/{tool}/assembly.fasta"),
        bam=os.path.join(RESULTS_DIR, "mapping/metag/lr/{tool}/polishing/assembly_lr.bam")
    output:
        fasta=os.path.join(RESULTS_DIR, "assembly/lr/{tool}/polishing/racon/assembly.fasta"),
        # NOTE: Tmp files for input data to pass file extension checks
        sam=temp(os.path.join(RESULTS_DIR, "assembly/lr/{tool}/polishing/racon/input.sam")),
        asm=temp(os.path.join(RESULTS_DIR, "assembly/lr/{tool}/polishing/racon/input.fasta"))
    log:
        out="logs/racon.{tool}.out.log",
        err="logs/racon.{tool}.err.log"
    wildcard_constraints:
        tool="|".join(config["assemblers"]["lr"])
    threads:
        config["racon"]["threads"]
    conda:
        os.path.join(ENV_DIR, "racon.yaml")
    message:
        "Assembly: long reads: polishing w/ Racon"
    shell:
        "(date && "
        "samtools view -h {input.bam} > {output.sam} && "
        "cp {input.asm} {output.asm} && "
        "racon -t {threads} {input.lr} {output.sam} {output.asm} > {output.fasta} && "
        "date) 2> {log.err} > {log.out}"

rule polishing_lr_medaka:
    input: 
        lr=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.fastq.gz"),
        asm=os.path.join(RESULTS_DIR, "assembly/lr/{tool}/polishing/racon/assembly.fasta")
    output:
        os.path.join(RESULTS_DIR, "assembly/lr/{tool}/polishing/racon_medaka/consensus.fasta")
    log:
        out="logs/racon.medaka.{tool}.out.log",
        err="logs/racon.medaka.{tool}.err.log"
    wildcard_constraints:
        tool="|".join(config["assemblers"]["lr"])
    threads:
        config["medaka"]["threads"]
    conda:
        os.path.join(ENV_DIR, "medaka.yaml")
    message:
        "Assembly: long reads: polishing w/ Medaka"
    shell:
        "(date && "
        "medaka_consensus -i {input.lr} -d {input.asm} -o $(dirname {output}) -t {threads} -m {config[medaka][model]} && "
        # "cd $(dirname {output}) && ln -sf consensus.fasta $(basename {output}) && " # NOTE: flye already creates this file
        "date) 2> {log.err} > {log.out}"

localrules: assembly_lr_flye_link

rule assembly_lr_flye_link:
    input:
        os.path.join(RESULTS_DIR, "assembly/lr/{tool}/polishing/racon_medaka/consensus.fasta")
    output:
        os.path.join(RESULTS_DIR, "assembly/lr/{tool}/ASSEMBLY.fasta")
    shell:
        "ipath=$(realpath {input}) && cd $(dirname {output}) && ln -sf $(realpath --relative-to=\".\" \"${{ipath}}\") $(basename {output})"

# Short reads
rule assembly_sr_megahit:
    input: 
        r1=os.path.join(RESULTS_DIR, "preproc/metag/sr/R1.fastp.fastq.gz"),
        r2=os.path.join(RESULTS_DIR, "preproc/metag/sr/R2.fastp.fastq.gz")
    output:
        os.path.join(RESULTS_DIR, "assembly/sr/megahit/ASSEMBLY.fasta")
    log:
        out="logs/megahit.out.log",
        err="logs/megahit.err.log"
    threads:
        config["megahit"]["threads"]
    conda:
        os.path.join(ENV_DIR, "megahit.yaml")
    message:
        "Assembly: short reads: MEGAHIT"
    shell:
        "(date && megahit -1 {input.r1} -2 {input.r2} -t {threads} -o $(dirname {output})/tmp && "
        "cd $(dirname {output}) && "
        "rsync -avP tmp/ . && "
        "ln -sf final.contigs.fa $(basename {output}) && "
        "rm -rf tmp/ && "
        "date) 2> {log.err} > {log.out}"

rule assembly_sr_metaspades:
    input:
        r1=os.path.join(RESULTS_DIR, "preproc/metag/sr/R1.fastp.fastq.gz"),
        r2=os.path.join(RESULTS_DIR, "preproc/metag/sr/R2.fastp.fastq.gz")
    output:
        os.path.join(RESULTS_DIR, "assembly/sr/metaspades/ASSEMBLY.fasta")
    log:
        out="logs/metaspades.out.log",
        err="logs/metaspades.err.log"
    threads:
        config["metaspades"]["threads"]
    conda:
        os.path.join(ENV_DIR, "spades.yaml")
    message:
        "Assembly: short reads: MetaSPAdes"
    shell:
        "(date && "
        "metaspades.py -k 21,33,55,77 -t {threads} -1 {input.r1} -2 {input.r2} -o $(dirname {output}) && "
        "cd $(dirname {output}) && ln -sf contigs.fasta $(basename {output}) && "
        "date) 2> {log.err} > {log.out}"

# Hybrid
rule assembly_hy_metaspades:
    input: 
        lr=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.fastq.gz"),
        r1=os.path.join(RESULTS_DIR, "preproc/metag/sr/R1.fastp.fastq.gz"),
        r2=os.path.join(RESULTS_DIR, "preproc/metag/sr/R2.fastp.fastq.gz")
    output: 
        os.path.join(RESULTS_DIR, "assembly/hy/metaspadeshybrid/ASSEMBLY.fasta")
    log:
        out="logs/metaspadeshybrid.out.log",
        err="logs/metaspadeshybrid.err.log"
    threads:
        config["metaspades"]["threads"]
    conda:
        os.path.join(ENV_DIR, "spades.yaml")
    message:
        "Assembly: hybrid: MetaSPAdes"
    shell:
        "(date && "
        "spades.py --meta -k 21,33,55,77 -t {threads} -1 {input.r1} -2 {input.r2} --nanopore {input.lr} -o $(dirname {output}) && "
        "cd $(dirname {output}) && ln -sf contigs.fasta $(basename {output}) && "
        "date) 2> {log.err} > {log.out}"

rule assembly_hy_operams:
    input: 
        lr=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.fastq.gz"),
        r1=os.path.join(RESULTS_DIR, "preproc/metag/sr/R1.fastp.fastq.gz"),
        r2=os.path.join(RESULTS_DIR, "preproc/metag/sr/R2.fastp.fastq.gz"),
        asm=os.path.join(RESULTS_DIR, "assembly/sr/megahit/ASSEMBLY.fasta")
    output:
        asm=os.path.join(RESULTS_DIR, "assembly/hy/operams/ASSEMBLY.fasta"),
        lr=temp(os.path.join(RESULTS_DIR, "assembly/hy/operams/lr.fastq"))
    log:
        out="logs/operams.out.log",
        err="logs/operams.err.log"
    threads:
        config["operams"]["threads"]
    conda:
        os.path.join(ENV_DIR, "operams.yaml")
    message:
        "Assembly: hybrid: OPERA-MS"
    shell:
       "(date && "
       "zcat {input.lr} > {output.lr} && "
       "perl {config[operams][bin]} --short-read1 {input.r1} --short-read2 {input.r2} --long-read {output.lr} --contig-file {input.asm} "
       "--out-dir $(dirname {output.asm}) --long-read-mapper minimap2 --num-processors {threads} && "
       "cd $(dirname {output.asm}) && ln -sf contigs.fasta $(basename {output.asm}) && "
       "date) 2> {log.err} > {log.out}"