# Mapping (to assembly)

##################################################
# For polishing

rule mapping_bwa_idx_polishing:
    input:
        os.path.join(RESULTS_DIR, "assembly/lr/{tool}/assembly.fasta")
    output:
        temp(expand(os.path.join(RESULTS_DIR, "assembly/lr/{{tool}}/polishing/assembly.{ext}"), ext=BWA_IDX_EXT))
    log:
        out="logs/bwa_index.polishing.{tool}.out.log",
        err="logs/bwa_index.polishing.{tool}.err.log"
    wildcard_constraints:
        tool="|".join(config["assemblers"]["lr"])
    threads: 1
    params:
        idx_prefix=lambda wildcards, output: os.path.splitext(output[0])[0]
    conda:
        os.path.join(ENV_DIR, "bwa.yaml")
    message:
        "Mapping: BWA index for assembly polishing"
    shell:
        "(date && bwa index {input} -p {params.idx_prefix} && date) 2> {log.err} > {log.out}"

rule mapping_bwa_mem_polishing:
    input:
        lr=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.fastq.gz"),
        asm=os.path.join(RESULTS_DIR, "assembly/lr/{tool}/assembly.fasta"),
        idx=expand(os.path.join(RESULTS_DIR, "assembly/lr/{{tool}}/polishing/assembly.{ext}"), ext=BWA_IDX_EXT)
    output:
        temp(os.path.join(RESULTS_DIR, "mapping/metag/lr/{tool}/polishing/assembly_lr.bam"))
    log:
        out="logs/bwa_mem.polishing.metag.lr.{tool}.out.log",
        err="logs/bwa_mem.polishing.metag.lr.{tool}.err.log"
    wildcard_constraints:
        tool="|".join(config["assemblers"]["lr"])
    threads:
        config["bwa"]["threads"]
    params:
        idx_prefix=lambda wildcards, input: os.path.splitext(input.idx[0])[0],
        bam_prefix=lambda wildcards, output: os.path.splitext(output[0])[0]
    conda:
        os.path.join(ENV_DIR, "bwa.yaml")
    message:
        "Mapping long reads to assembly w/ BWA for polishing"
    shell:
        "(date && "
        "bwa mem -x ont2d -t {threads} {params.idx_prefix} {input.lr} | "
        "samtools view -@ {threads} -SbT {input.asm} | "
        "samtools sort -@ {threads} -m {config[samtools][sort][chunk_size]} -T {params.bam_prefix} > {output} && "
        "date) 2> {log.err} > {log.out}"

##################################################
# For assembly

# BWA index
rule mapping_bwa_idx_assembly:
    input:
        os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta")
    output:
        expand(os.path.join(RESULTS_DIR, "assembly/{{rtype}}/{{tool}}/ASSEMBLY.{ext}"), ext=BWA_IDX_EXT)
    log:
        out="logs/bwa_index.{rtype}.{tool}.out.log",
        err="logs/bwa_index.{rtype}.{tool}.err.log"
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads: 1
    params:
        idx_prefix=lambda wildcards, output: os.path.splitext(output[0])[0]
    conda:
        os.path.join(ENV_DIR, "bwa.yaml")
    message:
        "Mapping: BWA index for assembly mapping"
    shell:
        "(date && bwa index {input} -p {params.idx_prefix} && date) 2> {log.err} > {log.out}"

# Short reads
rule mapping_bwa_mem_assembly_sr:
    input:
        # reads
        r1=os.path.join(RESULTS_DIR, "preproc/{mtype}/sr/R1.fastp.fastq.gz"),
        r2=os.path.join(RESULTS_DIR, "preproc/{mtype}/sr/R2.fastp.fastq.gz"),
        # assembly
        asm=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta"),
        # assembly index
        idx=expand(os.path.join(RESULTS_DIR, "assembly/{{rtype}}/{{tool}}/ASSEMBLY.{ext}"), ext=BWA_IDX_EXT)
    output:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype}/{tool}/ASSEMBLY_sr.bam")
    log:
        out="logs/bwa_mem.{mtype}.{rtype}.{tool}.sr.out.log",
        err="logs/bwa_mem.{mtype}.{rtype}.{tool}.sr.err.log"
    wildcard_constraints:
        mtype="|".join(META_TYPES),
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads:
        config["bwa"]["threads"]
    params:
        idx_prefix=lambda wildcards, input: os.path.splitext(input.idx[0])[0],
        bam_prefix=lambda wildcards, output: os.path.splitext(output[0])[0]
    conda:
        os.path.join(ENV_DIR, "bwa.yaml")
    message:
        "Mapping short reads to assembly w/ BWA"
    shell:
        "(date && "
        "bwa mem -t {threads} {params.idx_prefix} {input.r1} {input.r2} | "
        "samtools view -@ {threads} -SbT {input.asm} | "
        "samtools sort -@ {threads} -m {config[samtools][sort][chunk_size]} -T {params.bam_prefix} > {output} && "
        "date) 2> {log.err} > {log.out}"

# Long reads
rule mapping_bwa_mem_assembly_lr:
    input:
        # reads
        lr=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.fastq.gz"),
        # assembly
        asm=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta"),
        # assembly index
        idx=expand(os.path.join(RESULTS_DIR, "assembly/{{rtype}}/{{tool}}/ASSEMBLY.{ext}"), ext=BWA_IDX_EXT)
    output:
        os.path.join(RESULTS_DIR, "mapping/metag/{rtype}/{tool}/ASSEMBLY_lr.bam")
    log:
        out="logs/bwa_mem.metag.{rtype}.{tool}.lr.out.log",
        err="logs/bwa_mem.metag.{rtype}.{tool}.lr.err.log"
    wildcard_constraints:
        rtype="|".join(["lr", "hy"]),
        tool="|".join(config["assemblers"]["lr"] + config["assemblers"]["hy"])
    threads:
        config["bwa"]["threads"]
    params:
        idx_prefix=lambda wildcards, input: os.path.splitext(input.idx[0])[0],
        bam_prefix=lambda wildcards, output: os.path.splitext(output[0])[0]
    conda:
        os.path.join(ENV_DIR, "bwa.yaml")
    message:
        "Mapping long reads to assembly w/ BWA"
    shell:
        "(date && "
        "bwa mem -x ont2d -t {threads} {params.idx_prefix} {input.lr} | "
        "samtools view -@ {threads} -SbT {input.asm} | "
        "samtools sort -@ {threads} -m {config[samtools][sort][chunk_size]} -T {params.bam_prefix} > {output} && "
        "date) 2> {log.err} > {log.out}"

# Hybrid: short + long reads
rule mapping_bwa_mem_assembly_hy:
    input:
        sr=os.path.join(RESULTS_DIR, "mapping/metag/hy/{tool}/ASSEMBLY_sr.bam"),
        lr=os.path.join(RESULTS_DIR, "mapping/metag/hy/{tool}/ASSEMBLY_lr.bam")
    output:
        os.path.join(RESULTS_DIR, "mapping/metag/hy/{tool}/ASSEMBLY_hy.bam")
    log:
        out="logs/bwa_mem.metag.hy.{tool}.hy.out.log",
        err="logs/bwa_mem.metag.hy.{tool}.hy.err.log"
    wildcard_constraints:
        tool="|".join(config["assemblers"]["hy"])
    threads: 1
    conda:
        os.path.join(ENV_DIR, "bwa.yaml")
    message:
        "Mapping: merging short-reads and long-reads mapping results"
    shell:
        "(date && samtools merge {output} {input.sr} {input.lr} && date) 2> {log.err} > {log.out}"

##################################################
# Assembly coverage

# Compute the genome coverage histogram - https://bedtools.readthedocs.io/en/latest/content/tools/genomecov.html
rule mapping_assembly_genomecov:
    input:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype1}/{tool}/ASSEMBLY_{rtype2}.bam")
    output:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype1}/{tool}/ASSEMBLY_{rtype2}.cov.txt")
    log:
        out="logs/bedtools_genomecov.{mtype}.{rtype1}.{tool}.{rtype2}.out.log",
        err="logs/bedtools_genomecov.{mtype}.{rtype1}.{tool}.{rtype2}.err.log"
    wildcard_constraints:
        mtype="|".join(META_TYPES),
        rtype1="|".join(READ_TYPES),
        rtype2="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads: 1
    conda:
        os.path.join(ENV_DIR, "bedtools.yaml")
    message:
        "Mapping: compute assembly coverage"
    shell:
        "(date && bedtools genomecov -ibam {input} > {output} && date) 2> {log.err} > {log.out}"

rule mapping_assembly_genomecov_average:
    input:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype1}/{tool}/ASSEMBLY_{rtype2}.cov.txt")
    output:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype1}/{tool}/ASSEMBLY_{rtype2}.cov.ave.txt")
    log:
        out="logs/ave_genomecov.{mtype}.{rtype1}.{tool}.{rtype2}.out.log",
        err="logs/ave_genomecov.{mtype}.{rtype1}.{tool}.{rtype2}.err.log"
    wildcard_constraints:
        mtype="|".join(META_TYPES),
        rtype1="|".join(READ_TYPES),
        rtype2="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads: 1
    params:
        script=os.path.join(SRC_DIR, "coverage.awk")
    message:
        "Mapping: compute average assembly coverage"
    shell:
        "(date && cat {input} | awk -f {params.script} | tail -n+2 > {output} && date) 2> {log.err} > {log.out}"

##################################################
# Mappability index

# overall mapped reads, unmapped reads etc. 
rule mapping_assembly_flagstat:
    input:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype1}/{tool}/ASSEMBLY_{rtype2}.bam")
    output:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype1}/{tool}/ASSEMBLY_{rtype2}.flagstat.txt")
    log:
        out="logs/ave_genomecov.{mtype}.{rtype1}.{tool}.{rtype2}.out.log",
        err="logs/ave_genomecov.{mtype}.{rtype1}.{tool}.{rtype2}.err.log"
    wildcard_constraints:
        mtype="|".join(META_TYPES),
        rtype1="|".join(READ_TYPES),
        rtype2="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    conda:
        os.path.join(ENV_DIR, "bwa.yaml")
    message:
        "Mapping: assembly coverage stats w/ samtools flagstat"
    shell:
        "(date && samtools flagstat {input} > {output} && date) 2> {log.err} > {log.out}"

# per contig information (used this to subset contigs by length)
rule mapping_assembly_idxstats:
    input:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype1}/{tool}/ASSEMBLY_{rtype2}.bam")
    output:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype1}/{tool}/ASSEMBLY_{rtype2}.idxstats.txt")
    log:
        out="logs/ave_genomecov.{mtype}.{rtype1}.{tool}.{rtype2}.out.log",
        err="logs/ave_genomecov.{mtype}.{rtype1}.{tool}.{rtype2}.err.log"
    wildcard_constraints:
        mtype="|".join(META_TYPES),
        rtype1="|".join(READ_TYPES),
        rtype2="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    conda:
        os.path.join(ENV_DIR, "bwa.yaml")
    message:
        "Mapping: assembly coverage stats w/ samtools idxstats"
    shell:
        "(date && samtools idxstats {input} > {output} && date) 2> {log.err} > {log.out}"

# number of uniquely mapped reads
rule mapping_assembly_uniq:
    input:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype1}/{tool}/ASSEMBLY_{rtype2}.bam")
    output:
        os.path.join(RESULTS_DIR, "mapping/{mtype}/{rtype1}/{tool}/ASSEMBLY_{rtype2}.unique.txt")
    log:
        out="logs/ave_genomecov.{mtype}.{rtype1}.{tool}.{rtype2}.out.log",
        err="logs/ave_genomecov.{mtype}.{rtype1}.{tool}.{rtype2}.err.log"
    wildcard_constraints:
        mtype="|".join(META_TYPES),
        rtype1="|".join(READ_TYPES),
        rtype2="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    conda:
        os.path.join(ENV_DIR, "bwa.yaml")
    message:
        "Mapping: assembly coverage stats: number of uniquely mapped reads"
    message:
        "Assembly mapping: unique: {input}"
    shell:
        "(date && samtools view {input} | grep -v \"XT:A:U\" | uniq | wc -l > {output} && date) 2> {log.err} > {log.out}"
