##############################
# Preprocessing

rule collect_nanostats:
    input:
        os.path.join(RESULTS_DIR, "qc/metag/lr/NanoStats.txt")
    output:
        os.path.join(RESULTS_DIR, "qc/metag/lr/NanoStats.tsv")
    run:
        import pandas
        from scripts.utils import parse_nanostats_report
        summary = parse_nanostats_report(input[0])
        summary.to_csv(output[0], sep="\t", header=True, index=False, index_label=False)

##############################
# Annotation

rule collect_plasflow:
    input:
        expand(
            os.path.join(RESULTS_DIR, "annotation/plasflow/{rtype_tool}/plasflow.tsv"),
            rtype_tool=["%s/%s" % (rtype, tool) for rtype, tool in READ_ASSEMBLERS]
        )
    output:
        os.path.join(RESULTS_DIR, "annotation/plasflow/summary.tsv")
    run:
        import pandas
        from scripts.utils import parse_plasflow_report

        # read in and summarize
        summary = []
        for ifile in input:
            ifile_rtype = os.path.basename(os.path.dirname(os.path.dirname(ifile)))
            ifile_tool  = os.path.basename(os.path.dirname(ifile))
            idf_sum = parse_plasflow_report(ifile)
            idf_sum["tool"] = ifile_tool
            summary.append(idf_sum)
        # concat and save
        summary = pandas.concat(objs=summary, axis="index")
        summary.to_csv(output[0], sep="\t", header=True, index=True, index_label="label")

rule collect_rgi:
    input:
        expand(
            os.path.join(RESULTS_DIR, "annotation/rgi/{rtype_tool}/rgi.txt"),
            rtype_tool=["%s/%s" % (rtype, tool) for rtype, tool in READ_ASSEMBLERS]
        )
    output:
        os.path.join(RESULTS_DIR, "annotation/rgi/summary.tsv")
    run:
        import pandas
        from scripts.utils import parse_rgi_report, merge_rgi_reports

        summary_types = ["all", "strict", "nudged"]
        summary_cols  = ["Best_Hit_ARO", "ARO", "Drug Class", "Resistance Mechanism", "AMR Gene Family"]
        
        # collect
        summaries = []
        for ifile in input:
            ifile_rtype = os.path.basename(os.path.dirname(os.path.dirname(ifile)))
            ifile_tool  = os.path.basename(os.path.dirname(ifile))
            summaries.append(parse_rgi_report(ifile, ifile_tool, summary_cols))
        # merge
        merged_summaries = []
        for summary_col in summary_cols:
            for summary_type in summary_types:
                merged_summaries.append(merge_rgi_reports(summaries, summary_col, summary_type))
        # concat
        summary = pandas.concat(objs=merged_summaries, axis="index")
        # reorder columns
        summary_cols_ordered = ["label", "col", "type"] + [c for c in summary.columns if c not in ["label", "col", "type"]]
        # save
        summary[summary_cols_ordered].to_csv(output[0], sep="\t", header=True, index=False, index_label=False, na_rep=0)

rule collect_casc:
    input:
        expand(
            os.path.join(RESULTS_DIR, "annotation/casc/{rtype_tool}/ASSEMBLY.results.txt"),
            rtype_tool=["%s/%s" % (rtype, tool) for rtype, tool in READ_ASSEMBLERS]
        )
    output:
        os.path.join(RESULTS_DIR, "annotation/casc/summary.tsv")
    run:
        import pandas
        from scripts.utils import parse_casc_report

        summary = []
        for ifile in input:
            ifile_rtype = os.path.basename(os.path.dirname(os.path.dirname(ifile)))
            ifile_tool  = os.path.basename(os.path.dirname(ifile))
            ifile_df = parse_casc_report(ifile)
            ifile_df = ifile_df.assign(crispr_tool="casc")
            ifile_df = ifile_df.assign(asm_tool=ifile_tool)
            summary.append(ifile_df)
        # concat and save
        summary = pandas.concat(objs=summary, axis="index")
        summary.to_csv(output[0], sep="\t", header=True, index=False, index_label=False)

rule collect_minced:
    input:
        expand(
            os.path.join(RESULTS_DIR, "annotation/minced/{rtype_tool}/ASSEMBLY.results.txt"),
            rtype_tool=["%s/%s" % (rtype, tool) for rtype, tool in READ_ASSEMBLERS]
        )
    output:
        os.path.join(RESULTS_DIR, "annotation/minced/summary.tsv")
    run:
        import pandas
        from scripts.utils import parse_minced_report

        summary = []
        for ifile in input:
            ifile_rtype = os.path.basename(os.path.dirname(os.path.dirname(ifile)))
            ifile_tool  = os.path.basename(os.path.dirname(ifile))
            ifile_df = parse_minced_report(ifile)
            ifile_df = ifile_df.assign(crispr_tool="minced")
            ifile_df = ifile_df.assign(asm_tool=ifile_tool)
            summary.append(ifile_df)
        # concat and save
        summary = pandas.concat(objs=summary, axis="index")
        summary.to_csv(output[0], sep="\t", header=True, index=False, index_label=False)

##############################
# Analysis

rule collect_quast:
    input:
        expand(
            os.path.join(RESULTS_DIR, "analysis/quast/{rtype_tool}/report.tsv"),
            rtype_tool=["%s/%s" % (rtype, tool) for rtype, tool in READ_ASSEMBLERS]
        )
    output:
        os.path.join(RESULTS_DIR, "analysis/quast/summary_report.tsv")
    run:
        import pandas
        summary = None
        for ifile in input:
            ifile_rtype = os.path.basename(os.path.dirname(os.path.dirname(ifile)))
            ifile_tool  = os.path.basename(os.path.dirname(ifile))
            ifile_df    = pandas.read_csv(ifile, sep="\t", header=0, index_col=0)
            ifile_df.rename(columns={"ASSEMBLY": ifile_tool}, inplace=True)
            print(ifile_df)
            if summary is None:
                summary = ifile_df.copy()
            else:
                summary = summary.merge(
                    right=ifile_df,
                    how="outer",
                    left_index=True,
                    right_index=True,
                    sort=False
                )
        summary.to_csv(output[0], sep="\t", header=True, index=True, index_label=False)

# NOTE: currently not used
# rule collect_samtools_flagstat:
#     input:
#         metag=expand(
#             os.path.join(RESULTS_DIR, "mapping/metag/{rtype_tool}.flagstat.txt"),
#             rtype_tool=["{r}/{t}/ASSEMBLY_{r}".format(r=r, t=t) for r, t in READ_ASSEMBLERS]
#         ),
#         metat=expand(
#             os.path.join(RESULTS_DIR, "mapping/metat/{rtype_tool}.flagstat.txt"),
#             rtype_tool=["{r}/{t}/ASSEMBLY_sr".format(r=r, t=t) for r, t in READ_ASSEMBLERS]
#         ) if "metat" in META_TYPES else []
#     output:
#         metag=os.path.join(RESULTS_DIR, "mapping/metag/mappability.tsv"),
#         metat=os.path.join(RESULTS_DIR, "mapping/metat/mappability.tsv")
#     run:
#         import pandas
#         from scripts.utils import parse_samtools_flagstat_report
#         def collect(ifiles):
#             summary = {}
#             for ifile in ifiles:
#                 ifile_rtype = os.path.basename(os.path.dirname(os.path.dirname(ifile)))
#                 ifile_tool  = os.path.basename(os.path.dirname(ifile))
#                 summary[ifile_tool] = parse_samtools_flagstat_report(ifile)
#             if len(summary) > 0:
#                 return pandas.DataFrame.from_dict(summary)
#             return None
#         metag_summary = collect(input.metag)
#         metat_summary = collect(input.metat)
#         metag_summary.to_csv(output.metag, sep="\t", header=True, index=True, index_label=False)
#         if metat_summary is not None:
#             metat_summary.to_csv(output.metat, sep="\t", header=True, index=True, index_label=False)
#         else:
#             from pathlib import Path
#             Path(output.metat).touch()

rule collect_samtools_idxstats:
    input:
        metag=expand(
            os.path.join(RESULTS_DIR, "mapping/metag/{rtype_tool}.idxstats.txt"),
            rtype_tool=["{r}/{t}/ASSEMBLY_{r}".format(r=r, t=t) for r, t in READ_ASSEMBLERS]
        ),
        metat=expand(
            os.path.join(RESULTS_DIR, "mapping/metat/{rtype_tool}.idxstats.txt"),
            rtype_tool=["{r}/{t}/ASSEMBLY_sr".format(r=r, t=t) for r, t in READ_ASSEMBLERS]
        ) if "metat" in META_TYPES else []
    output:
        metag=os.path.join(RESULTS_DIR, "mapping/metag/mappability.tsv"),
        metat=os.path.join(RESULTS_DIR, "mapping/metat/mappability.tsv")
    params:
        minlengths=[0, 1000, 2000, 5000]
    run:
        import pandas
        from scripts.utils import parse_samtools_idxstats_report

        def collect(ifiles):
            summary = []
            for ifile in ifiles:
                ifile_rtype = os.path.basename(os.path.dirname(os.path.dirname(ifile)))
                ifile_tool  = os.path.basename(os.path.dirname(ifile))
                ifile_df    = parse_samtools_idxstats_report(ifile, params.minlengths)
                ifile_df    = ifile_df.assign(tool=ifile_tool)
                summary.append(ifile_df)
            if len(summary) > 0:
                return  pandas.concat(objs=summary, axis="index")
            return None
        
        metag_summary = collect(input.metag)
        metat_summary = collect(input.metat)
        metag_summary.to_csv(output.metag, sep="\t", header=True, index=False, index_label=False, na_rep="NA")
        if metat_summary is not None:
            metat_summary.to_csv(output.metat, sep="\t", header=True, index=False, index_label=False, na_rep="NA")
        else:
            from pathlib import Path
            Path(output.metat).touch()

rule collect_diamond_sq_len_ratio:
    input:
        expand(
            os.path.join(RESULTS_DIR, "annotation/diamond/{rtype_tool}/proteins.tsv"),
            rtype_tool=["%s/%s" % (rtype, tool) for rtype, tool in READ_ASSEMBLERS]
        )
    output:
        os.path.join(RESULTS_DIR, "annotation/diamond/summary_qs_len_ratio.tsv")
    run:
        import pandas
        from scripts.utils import parse_diamond_tsv
        summary = []
        for ifile in input:
            ifile_rtype = os.path.basename(os.path.dirname(os.path.dirname(ifile)))
            ifile_tool  = os.path.basename(os.path.dirname(ifile))
            ifile_df    = parse_diamond_tsv(ifile)
            ifile_df    = ifile_df.assign(tool=ifile_tool)
            summary.append(ifile_df[["qs_len_ratio", "tool"]].copy())
        summary = pandas.concat(objs=summary, axis="index")
        summary.to_csv(output[0], sep="\t", header=True, index=False, index_label=False)

# Number of total/partial genes
rule collect_prodigal:
    input:
        expand(
            os.path.join(RESULTS_DIR, "annotation/prodigal/{rtype_tool}/proteins.faa"),
            rtype_tool=["%s/%s" % (rtype, tool) for rtype, tool in READ_ASSEMBLERS]
        )
    output:
        os.path.join(RESULTS_DIR, "annotation/prodigal/summary_gene_counts.txt")
    threads: 1
    shell:
        """
        echo -e \"tool\\ttotal\\tpartial\" > {output}
        for f in {input}
        do
            echo -e \"$(basename $(dirname ${{f}}))\\t$(grep '^>' ${{f}} | wc -l)\\t$(grep '^>' ${{f}} | grep 'partial=\(10\|01\|11\);' | wc -l)\"
        done >> {output}
        """

rule collect_cdhit:
    input:
        expand(
            os.path.join(RESULTS_DIR, "analysis/cdhit/{combi}.faa"),
            combi=["%s_%s__%s_%s" % (p[0][0], p[0][1], p[1][0], p[1][1]) for p in READ_ASSEMBLER_PAIRS]
        ),
        expand(
            os.path.join(RESULTS_DIR, "analysis/cdhit/{combi}.faa"),
            combi=["%s_%s__%s_%s" % (p[1][0], p[1][1], p[0][0], p[0][1]) for p in READ_ASSEMBLER_PAIRS]
        )
    output:
        os.path.join(RESULTS_DIR, "analysis/cdhit/summary.tsv")
    shell:
        """
        echo -e \"rtype1\\ttool1\\trtype2\\ttool2\\tunique\" > {output}
        for f in {input}
        do
            f_bname=$(basename -s '.faa' ${{f}})
            echo -e \"$(echo ${{f_bname}} | sed 's/__/\\t/' | sed 's/_/\\t/g')\\t$(grep '^>' ${{f}} | wc -l)\"
        done >> {output}
        """
